local nmap = require "nmap"
local bin = require "bin"
local ipOps = require "ipOps"
local packet = require "packet"
local stdnse = require "stdnse"
local nsedebug = require "nsedebug"
local match = require "match"

-- author = "Joel Spang"

_ENV = stdnse.module("ads", stdnse.seeall);

-- Constants
-- Sizes in bytes
SIZES 	= {
	PREAMBLE	= 4,
	INDEXGRP	= 8,
	NETID 		= 6,
	FILLER 		= 6,
	PORT 		= 2,
	UNKNOWN 	= 8,
	HOSTNAME	= 9,
	VERSION		= 4,
	OS 			= 301,
	ERROR_CODE 	= 4
}

function get_bytes(data, offset, len)
	return data:sub(offset, (offset-1) + len), offset + len
end

function decode_os(data)
	return "?"
end

local function generate_netid_from_ip(id)
	local p = {}
	for c in string.gmatch(id, "%d+") do
	  table.insert(p, tonumber(c))
	end
	return string.format(				
		string.rep("%02x", 6),
		p[1], p[2], p[3], p[4], p[5] or 1, p[6] or 1
	)
end

local function convert_to_hex(id, isNum)
	local hex_str = ""
	if isNum then
		hex_str = string.format("%0X", tonumber(id))
		if (string.len(hex_str) % 2 ~= 0) then
			hex_str = "0"..hex_str 
		end
		return hex_str
	else
		for c in string.gmatch(id, ".") do
			local x = string.format("%X", c:byte())
			if string.len(x) == 1 then
				x = "0"..x
			end
			hex_str = hex_str .. x
		end
		return hex_str
	end
end

function build_route_packet(net_id, username, password, route_name, isEncrypted)
	-- Build the data
	local preamble 			= "03661471"									-- 4 bytes
	local message_type 		= string.format("00000000%s0000%s", "06", "00") -- 8 bytes	
	local unknown_filler	= "102705".."000000".."0c".."00"		
	local route 			= convert_to_hex(string.len(route_name or "N") + 1, true).."00"..convert_to_hex(route_name or "N") .. "00"
	local filler2 			= "07000600"
	local encryption		= ((isEncrypted and "0f00") or "0d00")

	local hexUser, hexPassword
	local filler3, filler4
	if not isEncrypted then
		hexUser 			= convert_to_hex(username) .. "00"
		hexPassword			= convert_to_hex(password) .. "00"
		filler3 			= convert_to_hex(string.len(username)+1, true).."00"
		filler4 			= "0200"..convert_to_hex(string.len(password)+1, true).."00"
	else
		hexUser				= username
		hexPassword 		= password
		filler3 			= convert_to_hex(string.len(username)/2, true).."00"
		filler4				= "0e00"..convert_to_hex(string.len(password)/2, true).."00"
	end
	 
	local filler5 			= "0500"
	local ending 			= "0000"

	local p = preamble .. message_type .. net_id .. unknown_filler .. route .. filler2
	p 		= p .. net_id .. encryption .. filler3 .. hexUser .. filler4 .. hexPassword .. filler5 .. route

	return bin.pack("H", p)
end

function build_discovery_packet(net_id)
	-- Build the data
	local preamble 			= "03661471"					-- 4 bytes
	local message_type 		= string.format("00000000%s0000%s", "01", "00") -- 8 bytes	
	local unknown_filler	= "0b14" .. string.rep("00", 4) -- 6 bytes

	return bin.pack("H", preamble .. message_type .. net_id .. unknown_filler)
end

function send_discover(host, port)
	local result, socket, try, catch
	local _, src_ip, src_port, dest_ip, dest_port
	
	if type(port) == "table" then
		port = port.number
	end

	socket = nmap.new_socket("udp")
	catch = function() 
		socket:close() 
	end

	try = nmap.new_try(catch)

	try(socket:bind(nil, port))
	try(socket:set_timeout(5000))
	try(socket:connect(host, port, "udp"))

	src_ip, src_port, dest_ip, dest_port = try(socket:get_info())

	-- Build our ADS packet
	local net_id = generate_netid_from_ip(src_ip)
	local packet_ads = build_discovery_packet(net_id)

	local x = try(socket:send(packet_ads)) 

	while( true ) do
		local status, response = socket:receive()

		if ( status == nil) then
			stdnse.print_debug(2, "Status is nil")
			break
		end

		-- Dissect the response
		local offset = 1
		local resp_preamble, offset 	= get_bytes(response, offset, SIZES.PREAMBLE)
		local resp_index_group, offset	= get_bytes(response, offset, SIZES.INDEXGRP)
		local resp_net_id, offset		= get_bytes(response, offset, SIZES.NETID)
		local resp_ads_port, offset		= get_bytes(response, offset, SIZES.PORT)
		local resp_unknown_1, offset	= get_bytes(response, offset, SIZES.UNKNOWN)
		local resp_name, offset			= get_bytes(response, offset, SIZES.HOSTNAME)
		local resp_os, offset			= get_bytes(response, offset, SIZES.OS)
		local resp_version, offset 		= get_bytes(response, offset, SIZES.VERSION)

		-- Generate net id (ip + ".1.1")
		local netid = ipOps.str_to_ip(resp_net_id:sub(1,4)) .. ".1.1"

		-- Convert ADS port from hex to dec
		local p = resp_ads_port
		local hex_port = string.format(
			"%2x%2x", string.byte(p:sub(1,1)), string.byte(p:sub(2,2))
		)
		local dec_port = tonumber(hex_port, 16)
		local str_os = decode_os(resp_os)

		-- Don't need the socket any longer
		socket:close()
		
		return {
			ip 		= host.ip,
			port 	= port,
			name 	= resp_name,
			netid 	= netid,
			adsport = dec_port,
			os 		= str_os
		}
	end

	return false
end

function send_route_establish(host, port, username, password, route, is_encrypted, net_id, socket)
	local result, socket, try, catch
	local _, src_ip, src_port, dest_ip, dest_port

	if type(port) == "table" then
		port = port.number
	end

	local socket = socket
	if socket == nil then
		socket = nmap.new_socket("udp")
		catch = function() 
			socket:close() 
		end

		try = nmap.new_try(catch)

		try(socket:bind(nil, port))
		try(socket:set_timeout(10))
		try(socket:connect(host, port, "udp"))
	end

	src_ip, src_port, dest_ip, dest_port = try(socket:get_info())

	-- Build our ADS packet	
	local _, src_ip, src_port, dest_ip, dest_port = socket:get_info()
	local net_id = generate_netid_from_ip(net_id or src_ip)
	local packet_ads = build_route_packet(net_id, username, password, route, is_encrypted)

	if ( not(socket:sendto(host, port, packet_ads)) ) then return false end

	local status, response = socket:receive_bytes(74)

	-- Parse the response and see if the route has been established
	local error_code = get_bytes(response, (string.len(response)) - SIZES.ERROR_CODE, SIZES.ERROR_CODE)
	local e = ""
	for c in string.gmatch(error_code, ".") do
		e = e .. c:byte()
	end

	return e and e == "0000" 
end

--[[function send_replay(host, port, username, password, encrypted, route, netid) 
	local socket = nmap.new_socket("udp")
	socket:set_timeout(1000)
		
	if type(port) == "table" then
		port = port.number
	end

	socket:bind(nil, 45666)

	local _, src_ip, src_port, dest_ip, dest_port = socket:get_info()

	-- Build our ADS packet
	local net_id = generate_netid_from_ip("192.168.1.66")
	local packet_ads = build_route_packet(net_id, username, password, route, encrypted)

	if ( not(socket:sendto(host, port, packet_ads)) ) then return false end

	local status, response = socket:receive_bytes(74)

	-- Parse the response and see if the route has been established
	local error_code = get_bytes(response, (string.len(response)) - SIZES.ERROR_CODE, SIZES.ERROR_CODE)
	local e = ""
	for c in string.gmatch(error_code, ".") do
		e = e .. c:byte()
	end

	return e and e == "0000" 
end]]


return _ENV;